/*
 *    Copyright (C) 2002 Sergej Bogomolov <bog@deol.ru>
 *    This filter cut range from YUV4MPEG2
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <yuv4mpeg.h>
#include <mjpeg_logging.h>

int verbose = 1;
int subtract = 0;
uint8_t	*input_frame[3];
uint8_t	*output_frame[3];
uint8_t *subtract_frame[3];

int	start_frame = 0;
int	end_frame = 999999999;

void	filter(int width, int height, uint8_t *input[], uint8_t * const output[]);

void
filter(int width, int height, uint8_t *input[], uint8_t * const output[])
{
    int i;
    uint8_t	*in, *out, *sub;
    
    if( subtract ) {
	for( i=0, in=input[0], out=output[0], sub=subtract_frame[0]; 
			i<width*height; i++ ) *out++ = *in++ - *sub++;
	for( i=0, in=input[1], out=output[1], sub=subtract_frame[1]; 
			i<width*height/4; i++ ) *out++ = *in++ - *sub++;
	for( i=0, in=input[2], out=output[2], sub=subtract_frame[2]; 
			i<width*height/4; i++ ) *out++ = *in++ - *sub++;
    } else {
	memcpy( output[0], input[0], width*height );
	memcpy( output[1], input[1], width*height/4 );
	memcpy( output[2], input[2], width*height/4 );
    }
}

size_t readbytes( int fd, void *buf, size_t count ) {
        ssize_t status;
        int rem = count;
        void *address=buf;
        while( rem > 0 ) {
                status = read( fd, address, rem );
                if( status <= 0 ) {
                        fprintf( stderr, "read status: %d\n", status );
                        return status;
                }
                rem -= status;
                address += status;
        }
        return count;
}

size_t readnl( int fd, void *buf ) {
        ssize_t status;
        int count;
        unsigned char *address=buf;
        for( count=0;;count++) {
                status = read( fd, address, 1 );
                if( status <= 0 ) return status;
                if( *address++ == '\n' ) return count;
        }
}

static void Usage(char *name )
{
	fprintf(stderr,
		"Usage: %s: [-h] [-s start_frame] [-e end_frame] [-v num]\n"
                        "-h   - Print out this help\n"
			"-s   - Start from this frame (default: 0)\n"
			"-e   - End with this frame (default: 999999999)\n"
			"-d filename - subtract from source file\n"
			"-v   - Verbosity [0..2]\n", name);
}
			
int
main(int argc, char *argv[])
{
	int	input_fd = 0;
	int	output_fd = 1;
	int	subtract_fd;
	int	horz;
	int	vert;
	int	c;
	int	frame_count;
	int i;
	char    header[128];
	int 	swidth, sheight;

	y4m_stream_info_t istream, ostream;
	y4m_frame_info_t iframe;

	while((c = getopt(argc, argv, "s:e:v:hd:")) != EOF) {
		switch(c) {
			case 's':
				start_frame = atoi(optarg);
				break;
			case 'e':
				end_frame = atoi(optarg);
				break;
		case 'v':
			verbose = atoi (optarg);
			if( verbose < 0 || verbose >2 )
			{
				Usage (argv[0]);
				exit (1);
			}
			break;
		case 'd':
			subtract_fd = open( optarg, O_RDONLY );
			if( subtract_fd == -1 ) 
				mjpeg_error_exit1("Subtract file error" );
			i = readbytes( subtract_fd, header, 10 );
			if( i != 10 ) { 
				perror( "Subtract file: input first string" ); 
				exit(1); 
			}
			if( strncmp( header, "YUV4MPEG2 ", 10 ) ) {
				perror( "Subtract file: no YUV4MPEG2 header" );
				exit(1); 
			}
			i = readnl( subtract_fd, header );
                	if( i <= 0 ) { 
			  perror( "Subtract file: input YUV4MPEG2 header" ); 
			  exit(1); 
			}
			sscanf( header, "W%d H%d", &swidth, &sheight );
			subtract = 1;
			break;
		case 'h':
                        Usage (argv[0]);
		default:
			exit(0);
		}
	}

	(void)mjpeg_default_handler_verbosity(verbose);

	y4m_init_stream_info(&istream);
	y4m_init_stream_info(&ostream);
	y4m_init_frame_info(&iframe);

	i = y4m_read_stream_header(input_fd, &istream);
	if (i != Y4M_OK)
		mjpeg_error_exit1("Input stream error: %s", y4m_strerr(i));

	if( istream.height % 2 != 0 )
	{
		mjpeg_error_exit1("Input images have odd number of lines" );
	}
	if( istream.width % 2 != 0 )
	{
		mjpeg_error_exit1("Input images have odd number of pixels" );
	}
	horz = istream.width;
	vert = istream.height;
	if( subtract ) {
	  if( horz != swidth ) 
		mjpeg_error_exit1( "differ in input/subtract streams width" );
	  if( vert != sheight )
		mjpeg_error_exit1( "differ in input/subtract streams height" );
	  subtract_frame[0] = malloc(horz * vert);
	  subtract_frame[1] = malloc((horz / 2) * (vert / 2));
	  subtract_frame[2] = malloc((horz / 2) * (vert / 2));
	}
	
	mjpeg_debug( "start=%d end=%d width=%d height=%d", 
			start_frame, end_frame, horz, vert);

	y4m_copy_stream_info(&ostream, &istream);

	input_frame[0] = malloc(horz * vert);
	input_frame[1] = malloc((horz / 2) * (vert / 2));
	input_frame[2] = malloc((horz / 2) * (vert / 2));

	output_frame[0] = malloc(horz * vert);
	output_frame[1] = malloc((horz / 2) * (vert / 2));
	output_frame[2] = malloc((horz / 2) * (vert / 2));

	y4m_write_stream_header(output_fd, &ostream);

	frame_count = 0;
	while (y4m_read_frame(input_fd, &istream, &iframe, input_frame) == Y4M_OK)
	{ 		
		if( frame_count > end_frame ) break;
 		if( frame_count < start_frame ) { frame_count++; continue; }
		if( subtract ) {
		  i = readnl( subtract_fd, header );
		  if( i <= 0 ) {
		    perror("Subtract YUV4MPEG2 frame header");
		    exit(1); 
		  }
		  i = readbytes( subtract_fd, subtract_frame[0], horz*vert );
		  if( i != horz*vert ) break;
		  i = readbytes( subtract_fd, subtract_frame[1], horz*vert/4 );
		  if( i != horz*vert/4 ) break;
		  i = readbytes( subtract_fd, subtract_frame[2], horz*vert/4 );
		  if( i != horz*vert/4 ) break;
		}
		filter(horz, vert,  input_frame, output_frame);
		y4m_write_frame(output_fd, &ostream, &iframe, output_frame);
		mjpeg_debug( "frame=%d", frame_count );
		frame_count++;
	}

	y4m_fini_stream_info(&istream);
	y4m_fini_stream_info(&ostream);
	y4m_fini_frame_info(&iframe);
	if( subtract ) close( subtract_fd );
	exit(0);
}


